/* Ansteuerung serielle Conrad-Relaisplatine 8fa Best. Nr. 96 77 20
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*

Rolf Freitag 2005: Added select, tcflush, timeout with pthread, ioctl for exclusive mode,
                   time checks, packet length checks, check of the address, printing of
                   the firmware version, ...
2004-2-11

 */

/* #define DEBUG */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>              /* File control definitions */
#include <termios.h>            /* POSIX terminal control definitions */
#include <pthread.h>
#include <iso646.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <setjmp.h>

/*
 * Befehlssatz der Karte:
 * <x> = 'dont't care', in der Regel 0
 * I_XOR = Pruefinfo (XOR aller Werte, wird vom Programm berechnet)
 *
 * Kommando                        Kommandorahmen      Antwort
 * -------------------------------------------------------------------------
 * 0  NO OPERATION - keine Aktion  0 <Adr.> <x> I_XOR    255 <Adr.>  <x> I_XOR
 * 1  SETUP - Initialisierung      1 <Adr.> <x> I_XOR    254 <Adr.> <info> I_XOR
 * 2  GET PORT - Schaltzustand     2 <Adr.> <x> I_XOR    254 <Adr.> <info> I_XOR
 * 3  SET PORT - Relais schalten   3 <Adr.> <data> I_XOR 253 <Adr.> <data> I_XOR
 * 4  GET OPTION - Option lesen    4 <Adr.> <x> I_XOR    254 <Adr.> <info> I_XOR
 * 5  SET OPTION - Option setzen   5 <Adr.> <data> I_XOR 254 <Adr.> <data> I_XOR
 * -------------------------------------------------------------------------
 *
 * Das Init-Kommando initialisiert alle Karten einer Kette. Der Computer
 * erhaelt also bei N Karten N+1 Antworten! Dasselbe gilt fuer
 * das GET PORT-Kommando.
 * Bei den Optionen koennen nur "enable broadcast" und "block broadcast"
 * gesetzt werden:
 *   Wert  Broadcast  Broadcast
 *        ausfuehren  blockieren
 *  ---------------------------------------------
 *    0     nein         nein
 *    1      ja          nein   (Voreinstellung)
 *    2     nein          ja
 *    3      ja           ja    (nicht sinnvoll)
 *  ---------------------------------------------
 * Bei blockierten Brodcast wird bei INIT und GET PORT nur ein NOP an
 * die nachfolgenden Platinen gesendet.
 *
 * Ausgangsport ist normalerweise ttS0 bis ttyS3.
 * Der Port muss fuer den user oder die Gruppe, unter der
 * das Programm laeuft, Schreibberechtigung haben.
 * Die User des Programms koennen in die Gruppe von ttySx
 * aufgenommen werden (meist 'tty' oder 'dialout').
 * Gegebenfalls kann man auch das Programm der Gruppe von ttySx
 * zuordnen und das SetGroupId-Bit setzen.
 */

#define PORT 0

#define EXIT_ONLY_ON_SUCCESS

static int g_fd;                // global file descriptor
static long long int g_lli_time0, g_lli_time1, g_lli_time2, g_lli_time3;        // global times
static jmp_buf env;             // environment
static int fd;                  // file descriptor

// time in microseconds
inline signed long long int
get_time ()
{
  static struct timeval ti;
  static struct timezone tzp;

  gettimeofday (&ti, &tzp);
  return (ti.tv_usec + 1000000 * ((long long int) time (NULL)));
}


int
open_port (int port)
{
  /*
   * Oeffnet seriellen Port
   * Gibt das Filehandle zurueck oder -1 bei Fehler
   * der Parameter port muss 0, 1, 2 oder 3 sein
   *
   * RS232-Parameter
   * - 19200 baud
   * - 8 bits/byte
   * - no parity
   * - no handshake
   * - 1 stop bit
   */
  int fd, modelines = 0;;
  struct termios options;
  switch (port)
  {
    case 0:
      fd = open ("/dev/ttyS0", O_RDWR | O_NOCTTY | O_NDELAY);
      break;
    case 1:
      fd = open ("/dev/ttyS1", O_RDWR | O_NOCTTY | O_NDELAY);
      break;
    case 2:
      fd = open ("/dev/ttyS2", O_RDWR | O_NOCTTY | O_NDELAY);
      break;
    case 3:
      fd = open ("/dev/ttyS3", O_RDWR | O_NOCTTY | O_NDELAY);
      break;
    case 10:
      fd = open ("/dev/ttyUSB0", O_RDWR | O_NOCTTY | O_NDELAY);
      break;
    default:
      fd = -1;
      break;
  }
  if (fd >= 0)
  {
    /* get the current options */
    fcntl (fd, F_SETFL, 0);
    if (tcgetattr (fd, &options) != 0)
      return (-1);

    cfsetispeed (&options, B19200);     /* setze 19200 bps */
    cfsetospeed (&options, B19200);
    /* setze Optionen */
    options.c_cflag &= ~PARENB; /* kein Paritybit */
    options.c_cflag &= ~CSTOPB; /* 1 Stopbit */
    options.c_cflag &= ~CSIZE;  /* 8 Datenbits */
    options.c_cflag |= CS8;
    options.c_cflag |= (CLOCAL | CREAD);        /* CD-Signal ignorieren */
    /* Kein Echo, keine Steuerzeichen, keine Interrupts */
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
    options.c_oflag &= ~OPOST;  /* setze "raw" Input */
    options.c_cc[VMIN] = 0;     /* warten auf min. 0 Zeichen */
    options.c_cc[VTIME] = 10;   /* Timeout 10 Sekunden */
    tcflush (fd, TCIOFLUSH);
    if (-1 == tcsetattr (fd, TCSAFLUSH, &options) != 0)
    {
      fprintf (stderr, "ERROR: Could not set terminal attributes for port %d !\n", port);
      return (-1);
    }
    // Set IO-Control-Parameter and check the device.
    if (-1 == ioctl (fd, TIOCEXCL, &modelines)) // Put the tty into exclusive mode.
    {
      fprintf (stderr, "ERROR at ioctl TIOCEXCL on port %d!\n", port);
      perror ("ioctl()");
      return (-1);
    }
  }
  return (fd);
}


int
sndcmd (int fd, unsigned char command, unsigned char addr, unsigned char data)
{
  /*
   * Sendet Kommando an die Relaiskarte
   * die Berechnung der Pruefsumme erfolgt automatisch
   * Rueckgabe: 0 bei Erfolg, -1 bei Fehler
   */
  unsigned char wbuf[4];
  int i;

  wbuf[0] = command;
  wbuf[1] = addr;
  wbuf[2] = data;
  /* Pruefsumme */
  wbuf[3] = wbuf[0] ^ wbuf[1] ^ wbuf[2];
#ifdef DEBUG
  printf (" -> send: %d %d %d %d\n", wbuf[0], wbuf[1], wbuf[2], wbuf[3]);
#endif
  usleep (50000L);              /* 50 ms Pause, verhindert "Flattern" bei zu schnellem Wechsel */
  i = write (fd, wbuf, 4);
  g_lli_time1 = get_time (NULL);        // store send time
  tcflush (fd, TCOFLUSH);       // flush output buffer
  if (4 == i)
    return (0);
  printf ("ERROR: Could send only %d of 4 Bytes\n", i);
  return (-1);
}


int
rcvstat (int fd, unsigned char *answer, unsigned char *addr, unsigned char *data)
{
  /*
   * Empfaengt Status von der Relaiskarte
   * die Berechnung der Pruefsumme erfolgt automatisch
   * Rueckgabe: 0 bei Erfolg, -1 bei Fehler
   */
  unsigned char rbuf[4] = { 0 };
  int i_xor, i_retval, i;
  long long int lli_timediff0, lli_timediff1;
  struct timeval timeout;
  fd_set readfs;                // file descriptor set

  memset (&timeout, 0, sizeof (timeout));
  timeout.tv_sec = 2;           // seconds
  FD_ZERO (&readfs);
  FD_SET (fd, &readfs);
  i_retval = select (fd + 1, &readfs, NULL, NULL, &timeout);
  if (0 != i_retval)
  {
    if (i_retval == -1)
    {
      perror ("select()");
      tcflush (fd, TCIOFLUSH);  // flush buffers
      return (-1);
    }
    g_lli_time2 = get_time (NULL);      // store first receive time
    i = read (fd, rbuf, 4);
    g_lli_time3 = get_time (NULL);      // store last receive time
    lli_timediff0 = g_lli_time3 - g_lli_time0;  // first receive time - send time
    lli_timediff1 = g_lli_time3 - g_lli_time1;  // last receive time - send time
    if ((lli_timediff0 < 5000) or (lli_timediff1 > 25000))      // < 5 ms or > 25 ms
    {
      printf ("frame error\n");
      return (-1);              // impossible short time or extrem long time
    }
    i_xor = rbuf[0] ^ rbuf[1] ^ rbuf[2];
#ifdef DEBUG
    printf (" -> recv: %d %d %d %d\n", rbuf[0], rbuf[1], rbuf[2], rbuf[3]);
#endif
    *answer = rbuf[0];
    *addr = rbuf[1];
    *data = rbuf[2];
    if ((i_xor == rbuf[3]) and (4 == i) and (1 == *addr))       // checksum of 4 bytes and addr ok
      return (0);
    else
    {
      if (i_xor != rbuf[3])
        printf ("ERROR: XOR checksum is %x but must be %x.\n", i_xor, rbuf[3]);
      if (4 != i)
        printf ("ERROR: Could receive only %d of 4 Bytes.\n", i);
      if (1 != *addr)
        printf ("ERROR: Receive address is %d but should be 1, ignoring.\n", *addr);
      return (-1);
    }
  }                             // if (0 != i_retval)
  fprintf (stderr, "Connection timeout (select failed).\n");
  return (-1);
}


void
help (void)
{
  /* Hilfetext ausgeben */
  fprintf (stderr, "Ansteuerprogramm fuer die serielle Relaiskarte von Conrad\n");
  fprintf (stderr, "No Parameters!\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "Usage: relais <Parameter>\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "Parameter:\n");
  fprintf (stderr, "   -stat:  Status der Relais als Dezimalzahl\n");
  fprintf (stderr, "           Bit=1: Relais an, Bit=0: Relais aus\n");
  fprintf (stderr, "           keine weiteren Parameter moeglich\n");
  fprintf (stderr, "   -off:   alle Relais aus\n");
  fprintf (stderr, "   -on:    alle Relais an\n");
  fprintf (stderr, "   -sx:    Relais x einschalten (1 <= x <= 8)\n");
  fprintf (stderr, "   -rx:    Relais x ausschalten (1 <= x <= 8)\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "Beispiel:\n");
  fprintf (stderr, "   relais -off -s1 -s3: Relais 1 und 3 einschalten\n");
  fprintf (stderr, "   relais -s4 -r3:      Relais 4 ein- und 3 ausschalten\n");
  fprintf (stderr, "   relais -on -r6:      alle Relais ausser 6 einschalten\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "Rueckgabewert:\n");
  fprintf (stderr, "OK, Kontaktstellung: <Kontaktstellung dezimal>\n");
  fprintf (stderr, "z. B. 'OK, Kontaktstellung: 5' --> Relais 1 und 3 on\n");
  fprintf (stderr, "Bei Fehler wird FAIL: und der komplette Status\n");
  fprintf (stderr, "zurueckgegeben (Antwortcode Adresse Daten/Info).\n");
  fprintf (stderr, "\n");
}


// simple timeout
void *
func_timeout (void *threadid)
{
#ifdef EXIT_ONLY_ON_SUCCESS
restart:
#endif
  sleep (2);                    // 5 s timeout
  if (0 >= g_fd)                // device file could not be opened
  {
    fprintf (stderr, "ERROR: Timeout, device file could not be opened, exiting.\n");
#ifndef EXIT_ONLY_ON_SUCCESS
    close (fd);
    exit (-1);
#else
    goto restart;
#endif
  }
  pthread_exit (NULL);
}


int
main (int argc, char *argv[])
{
  static int i_ret, id;         /* File descriptor for the port, return value, thread id */
  static unsigned char ans, adr, stat, val, rval, n, retval;
  static pthread_t thread;

  (void) setjmp (env);
  // create thread for timeout (to avoid hangup in open_serial, e. g. when the device can not be opened)
  i_ret = pthread_create (&thread, NULL, func_timeout, (void *) id);    //(void *(*)(void *))
  if (i_ret)
  {
    printf ("ERROR; return code from pthread_create() is %d\n", i_ret);
    goto err_end;
  }

  fd = open_port (PORT);
  if (fd == -1)
  {
    fprintf (stderr, "Cannot open port %d\n", PORT);
    goto err_end;
  }
  else
    g_fd = fd;
  tcflush (fd, TCIOFLUSH);      // flush buffers
  g_lli_time0 = get_time (NULL);        // store start time

  if ((argc <= 1) || (strncmp (argv[1], "--help", 123) == 0))
  {
    help ();
    close (fd);
    pthread_cancel (thread);    // terminate timeout thread
    exit (0);
  }

  /* Karte Initialisieren */
  sndcmd (fd, 1, 1, 0);
  val = rcvstat (fd, &ans, &adr, &stat);
  if ((val != 0) || (ans != 254))
  {
    fprintf (stderr, "FAIL: card init failed\n");
    goto err_end;
  }
  printf ("Firmware-Version %hhu\n", stat);
  /* zweite Antwort vernichten */
  val = rcvstat (fd, &ans, &adr, &rval);

  /* Aktuellen Stand abfragen */
  do
  {
    sndcmd (fd, 2, 1, 0);
    val = rcvstat (fd, &ans, &adr, &rval);
  }
  while ((val == 0) && (ans != 253));

  /* Parameter auswerten */
  for (n = 1; n < argc; n++)
  {
    if (strncmp (argv[n], "-stat", 123) == 0)
    {
      /* Status liefern */
      if (ans == 253)
        printf ("OK, Kontaktstellung: %d\n", rval);
      else
      {
        printf ("FAIL: %d %d %d\n", ans, adr, rval);
        goto err_end;
      }
      close (fd);
      pthread_cancel (thread);  // terminate timeout thread
      exit (0);
    }
    else
    {
      if (strncmp (argv[n], "-off", 123) == 0)
      {
        /* Alles ausschalten */
        rval = 0;
      }
      else
      {
        if (strncmp (argv[n], "-on", 123) == 0)
        {
          /* Alles einschalten */
          rval = 255;
        }
        else
        {
          if ((argv[n][0] == '-') && (argv[n][1] == 's'))
          {
            val = atoi (&argv[n][2]);   /* Zahl hinter dem "-s" umwandeln */
            if ((1 <= val) && (val <= 8))
            {
              val = 1 << (val - 1);
              rval = rval | val;
            }
          }
          else
          {
            if ((argv[n][0] == '-') && (argv[n][1] == 'r'))
            {
              val = atoi (&argv[n][2]); /* Zahl hinter dem "-r" umwandeln */
              if ((1 <= val) && (val <= 8))
              {
                val = ~(1 << (val - 1));
                rval = rval & val;
              }
            }
            else
            {
              fprintf (stderr, "ERROR: Wrong Parameter: %s, exiting.\n", argv[n]);
              close (fd);
              pthread_cancel (thread);  // terminate timeout thread
              exit (-1);
            }
          }
        }
      }
    }
  }
  sndcmd (fd, 3, 1, rval);
  rcvstat (fd, &ans, &adr, &stat);
  close (fd);
  pthread_cancel (thread);      // terminate timeout thread
  if (ans != 252)
  {
    printf ("FAIL: %d %d %d\n", ans, adr, stat);
    goto err_end;
  }
  printf ("OK, Kontaktstellung: %d\n", stat);
  exit (0);
err_end:
  close (fd);
  pthread_cancel (thread);      // terminate timeout thread
  retval = -1;                  // use another value before goto and delete this line if you need different error codes
#ifdef EXIT_ONLY_ON_SUCCESS
  longjmp (env, retval);
#else
  exit (retval);
#endif
}
